# -*- coding: utf-8 -*-

import json
from tornado import stack_context

from . import constants
from .request import Request


class Connection(object):
    def __init__(self, server, stream, address, terminator=None, endpoint_name=None):
        self.server = server
        self.stream = stream
        self.address = address
        self.terminator = terminator or '\n'
        self.endpoint_name = endpoint_name or 'endpoint'

        self._clear_request_state()

        self.stream.set_close_callback(self._on_connection_close)
        # 开始等待数据
        self._read_message()

        # create_conn_func_list
        for func in self.server.create_conn_func_list:
            try:
                func(self)
            except Exception, e:
                self.server.logger.error('create_conn_func_list raise exception. func: %s, e: %s, traceback: %s',
                                         func.__name__, e, __import__('traceback').format_exc())

    def write(self, data, callback=None):
        """
        发送数据(实现是放到发送队列)
        """
        if isinstance(data, dict):
            data = self.make_rsp(**data)

        if not self.stream.closed():
            self._write_callback = stack_context.wrap(callback)
            self.stream.write(data, self._on_write_complete)

    def finish(self):
        """Finishes the request."""
        self._request_finished = True
        # No more data is coming, so instruct TCP to send any remaining
        # data immediately instead of waiting for a full packet or ack.
        self.stream.set_nodelay(True)
        if not self.stream.writing():
            self.close()

    def close(self, exc_info=False):
        """
        直接关闭连接
        注意: 不要在write完了之后直接调用，会导致write发送不出去
        """
        self.stream.close(exc_info)
        self._clear_request_state()

    def make_rsp(self, *args, **kwargs):
        """
        生成rsp
        """
        return json.dumps(dict(*args, **kwargs)) + self.terminator

    def set_close_callback(self, callback):
        """Sets a callback that will be run when the connection is closed.
        """
        self._close_callback = stack_context.wrap(callback)

    def _on_connection_close(self):
        # 链接被关闭的回调
        self.server.logger.debug('socket closed')

        if self._close_callback is not None:
            callback = self._close_callback
            self._close_callback = None
            callback()

        # close_conn_func_list
        for func in self.server.close_conn_func_list:
            try:
                func(self)
            except Exception, e:
                self.server.logger.error('close_conn_func_list raise exception. func: %s, e: %s, traceback: %s',
                                         func.__name__, e, __import__('traceback').format_exc())

    def _read_message(self):
        if not self.stream.closed():
            self.stream.read_until(self.terminator, self._on_read_complete)

    def _on_read_complete(self, raw_data):
        """
        数据获取结束
        """
        self.server.logger.debug('raw_data: %s', raw_data)
        request = Request(self, raw_data)
        self._handle_request(request)
        # 处理完了还是要继续保持可读
        self._read_message()

    def _on_write_complete(self):
        if self._write_callback is not None:
            callback = self._write_callback
            self._write_callback = None
            callback()
        if self._request_finished and not self.stream.writing():
            self.close()

    def _clear_request_state(self):
        """Clears the per-request state.

        This is run in between requests to allow the previous handler
        to be garbage collected (and prevent spurious close callbacks),
        and when the connection is closed (to break up cycles and
        facilitate garbage collection in cpython).
        """
        self._request_finished = False
        self._write_callback = None
        self._close_callback = None

    def _handle_request(self, request):
        """
        出现任何异常的时候，服务器不再主动关闭连接
        """

        try:
            request.json_data = json.loads(request.raw_data)
        except Exception, e:
            error = 'parse raw_data fail. e: %s, raw_data: %s' % (e, request.raw_data)
            self.server.logger.error(error)
            self.write(dict(ret=constants.RET_SYSTEM, error=error))
            return None

        if not isinstance(request.json_data, dict):
            error = 'json_data is not dict. json_data: %s' % request.json_data
            self.server.logger.error(error)
            self.write(dict(ret=constants.RET_SYSTEM, error=error))
            return None

        if not request.json_data.get(self.endpoint_name):
            error = 'endpoint is not in json_data. json_data: %s' % request.json_data
            self.server.logger.error(error)
            self.write(dict(ret=constants.RET_SYSTEM, error=error))
            return None

        endpoint = request.json_data[self.endpoint_name]
        view_func = self.server.get_route_view_func(endpoint)
        if not view_func:
            error = 'endpoint is not valid. endpoint: %s' % endpoint
            self.server.logger.error(error)
            self.write(dict(ret=constants.RET_SYSTEM, error=error))
            return None

        # before_request_func_list
        for func in self.server.before_request_func_list:
            try:
                func(request)
            except Exception, e:
                self.server.logger.error('before_request_func_list raise exception. func: %s, e: %s, traceback: %s',
                                         func.__name__, e, __import__('traceback').format_exc())
        view_func_exc = None
        view_func_result = None

        try:
            view_func_result = view_func(request)
        except Exception, e:
            error = 'view_func raise exception. endpoint: %s, view_func: %s, e: %s, traceback: %s' % (
                endpoint, view_func.__name__, e, __import__('traceback').format_exc())
            self.server.logger.error(error)
            view_func_exc = e
            self.write(dict(ret=constants.RET_SYSTEM,
                            error=error if self.server.debug else constants.ERROR_INTERNAL))

        # after_request_func_list
        for func in self.server.after_request_func_list:
            try:
                # 异常优先，以为一旦有异常，肯定没返回
                func(request, view_func_exc or view_func_result)
            except Exception, e:
                self.server.logger.error('after_request_func_list raise exception. func: %s, view_func_result: %s, e: %s, traceback: %s',
                                         func.__name__, view_func_result, e, __import__('traceback').format_exc())

        return view_func_result
